#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
import cv2
from tqdm import tqdm, trange
from collections import defaultdict
from scipy import signal
import queue
import threading

from aniposelib.cameras import CameraGroup

from .common import make_process_fun, get_nframes, \
    get_video_name, get_cam_name, \
    get_video_params, get_video_params_cap, \
    get_data_length, natural_keys, true_basename, find_calibration_folder

from .triangulate import load_offsets_dict

from .label_videos import label_frame

def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]

def write_frame_thread(writer, q):
    while True:
        frame = q.get(block=True)
        if frame is None:
            return
        writer.write(frame)
        # writer.writeFrame(frame)

def turn_to_black(frame):
    frame = np.float32(frame)
    white = np.all(frame > 220, axis=2)
    frame[white] = [0,0,0]
    frame[~white] *= 1.5
    frame = np.clip(frame, 0, 255).astype('uint8')
    return frame


def read_frames(caps_2d, cap_3d):
    frames_2d = []
    for cap in caps_2d:
        ret, frame = cap.read()
        if not ret:
            return False, None, None
        # img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        img = frame
        frames_2d.append(img)

    ret, frame = cap_3d.read()
    if not ret:
        return False, None, None
    frame_3d = frame
    # frame_3d = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    # frame_3d = turn_to_black(frame_3d)

    return ret, frames_2d, frame_3d

def get_video_params_cap(cap):
    params = dict()
    params['width'] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    params['height'] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    params['nframes'] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    params['fps'] = cap.get(cv2.CAP_PROP_FPS)
    return params

def draw_seq(img, seq, rect, range_y=(None, None),
             color=(0, 0, 0), thickness=5):
    low, high = range_y

    if low is None:
        low = np.min(seq)
    if high is None:
        high = np.max(seq)
    s = np.clip(seq, low, high)
    s = (s - low)/(high-low)

    left, right, top, bottom = rect
    height = bottom - top

    xs = np.linspace(left, right, num=len(seq))
    ys = (1-s) * height + top
    pointlist = list(zip(xs, ys))
    pointlist = [(x, y) for x, y in pointlist if not np.isnan(y)]
    pointlist = np.int32([pointlist])

    cv2.polylines(img, pointlist, False, color,
                  thickness=thickness, lineType=cv2.LINE_AA)

def mapto(x, fromLow, fromHigh, toLow, toHigh):
    norm = (x - fromLow) / (fromHigh-fromLow)
    return norm * (toHigh - toLow) + toLow

def draw_axis_y(img, rect, range_y, label,
                num_ticks=5,
                color=(0, 0, 0), thickness=5):
    left, right, top, bottom = rect
    height = bottom - top

    left_start = left - 10

    low, high = range_y
    ticks = np.linspace(low, high, num_ticks+2)[1:-1]

    cv2.line(img, (left_start, top+10), (left_start, bottom-10),
             color, thickness=thickness)

    font_face = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = cv2.getFontScaleFromHeight(
        font_face, int(round(height / (num_ticks + 3))))

    for t in ticks:
        y = mapto(t, low, high, bottom, top)
        y = int(round(y))
        lab = str(int(round(t)))
        (w, h), baseline = cv2.getTextSize(lab, font_face, font_scale, 2)
        cv2.line(img, (left_start-5, y), (left_start, y),
                 color, thickness=thickness)
        cv2.putText(img, lab, (left_start-30-w//2, y + baseline),
                    font_face, 0.9, color, thickness=2)


    imgnew = np.zeros((img.shape[1], img.shape[0]), dtype='uint8')
    cv2.putText(imgnew, label, (img.shape[0] - top - 130, left-100), font_face, font_scale*0.7, 255, thickness=2)
    imgnew = cv2.rotate(imgnew, cv2.ROTATE_90_COUNTERCLOCKWISE)

    img[imgnew > 0] = color

    return img


def get_plotting_params(caps_2d, cap_3d, ang_names=[]):

    height_angle = 175
    spacing_angle = 40
    spacing_videos = 20

    n_angles = len(ang_names)

    params_2d = [get_video_params_cap(c) for c in caps_2d]
    height_2d = max([p['height'] for p in params_2d])
    widths_2d = [round(p['width'] * height_2d/p['height']) for p in params_2d]

    param_3d = get_video_params_cap(cap_3d)
    height_3d = param_3d['height']
    width_3d = param_3d['width']

    start_3d = height_2d + spacing_videos
    start_angles = start_3d + height_3d + spacing_videos + spacing_angle

    width_total = sum(widths_2d)
    height_total = height_2d + spacing_videos + height_3d + spacing_videos + \
        height_angle * n_angles + spacing_angle*(n_angles+1)

    nframes = min([p['nframes'] for p in params_2d])
    nframes = min(nframes, param_3d['nframes'])

    fps = param_3d['fps']

    height_font = spacing_angle//2
    font_face = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = cv2.getFontScaleFromHeight(
        font_face, height_font)
    font_scale = int(round(font_scale))
    font_color = (0, 0, 0)
    font_thickness = 2

    mid_3d = int((width_total - width_3d) / 2)

    d = {
        'height_angle': height_angle,
        'spacing_angle': spacing_angle,
        'spacing_videos': spacing_videos,

        'height_2d': height_2d,
        'widths_2d': widths_2d,
        'start_3d': start_3d,
        'width_3d': width_3d,
        'mid_3d': mid_3d,
        'height_3d': height_3d,

        'nframes': nframes,
        'fps': fps,

        'width_total': width_total,
        'height_total': height_total,

        'start_3d': start_3d,
        'start_angles': start_angles,

        'height_font': height_font,
        'font_face': font_face,
        'font_scale': font_scale,
        'font_color': font_color,
        'font_thickness': font_thickness,
    }

    return d

def get_start_image(pp, ang_names=[]):

    start_img = np.zeros((pp['height_total'], pp['width_total'], 3), dtype='uint8')
    start_img[:] = 255

    for angnum, name in enumerate(ang_names):
        start_y = pp['start_angles'] + (pp['height_angle'] + pp['spacing_angle'])*angnum
        rect = (150, pp['width_total']-100, start_y, start_y + pp['height_angle'])

        font_size, baseline = cv2.getTextSize(
            name, pp['font_face'], pp['font_scale'], pp['font_thickness'])
        fw, fh = font_size

        text_xy = (pp['width_total'] // 2 - fw // 2, start_y)
        cv2.putText(start_img, name, text_xy, pp['font_face'], pp['font_scale'], pp['font_color'],
                    thickness=2, lineType=cv2.LINE_AA)

        draw_axis_y(start_img, rect, (0, 180), 'Angle',
                    num_ticks=3, thickness=4)

    return start_img


def draw_data(start_img, frames_2d, frame_3d, all_angles, pp):
    height_2d = pp['height_2d']
    widths_2d = pp['widths_2d']

    start_3d = pp['start_3d']
    height_3d = pp['height_3d']
    width_3d = pp['width_3d']
    mid_3d = pp['mid_3d']

    start_angles = pp['start_angles']
    height_angle = pp['height_angle']
    spacing_angle = pp['spacing_angle']

    width_total = pp['width_total']
    height_total = pp['height_total']

    frames_2d_resized = [cv2.resize(f, (w, height_2d))
                         for f, w in zip(frames_2d, widths_2d)]

    imout = np.copy(start_img)
    imout[0:height_2d] = np.hstack(frames_2d_resized)
    imout[start_3d:(start_3d + height_3d), mid_3d:(mid_3d+width_3d)] = frame_3d

    data_color = (0,0,0)
    indicator_color = (150,150,150)

    for angnum, angles in enumerate(all_angles):
        start_y = start_angles + (height_angle + spacing_angle)*angnum
        rect = (150, width_total-100, start_y, start_y + height_angle)
        left, right, top, bottom = rect

        draw_seq(imout, angles, rect,
                 range_y=(0, 180), color=data_color, thickness=2)
        x = (left+right)//2
        cv2.line(imout, (x, top+15), (x, bottom-15),
                 indicator_color, thickness=2)

    return imout

## TODO: remove this function and import from project_2d.py
def get_projected_points(config, pose_fname, cgroup, offsets_dict):
    try:
        scheme = config['labeling']['scheme']
    except KeyError:
        scheme = []

    pose_data = pd.read_csv(pose_fname)
    cols = [x for x in pose_data.columns if '_error' in x]
    if len(scheme) == 0:
        bodyparts = [c.replace('_error', '') for c in cols]
    else:
        bodyparts = sorted(set([x for dx in scheme for x in dx]))

    M = np.identity(3)
    center = np.zeros(3)
    for i in range(3):
        center[i] = np.mean(pose_data['center_{}'.format(i)])
        for j in range(3):
            M[i, j] = np.mean(pose_data['M_{}{}'.format(i, j)])

    bp_dict = dict(zip(bodyparts, range(len(bodyparts))))

    all_points = np.array([np.array(pose_data.loc[:, (bp+'_x', bp+'_y', bp+'_z')])
                           for bp in bodyparts])

    all_errors = np.array([np.array(pose_data.loc[:, bp+'_error'])
                           for bp in bodyparts])

    if config['triangulation']['optim']:
        all_errors[np.isnan(all_errors)] = 0
    else:
        all_errors[np.isnan(all_errors)] = 10000
    good = (all_errors < 100)
    all_points[~good] = np.nan

    n_joints, n_frames, _ = all_points.shape
    n_cams = len(cgroup.cameras)

    all_points_flat = all_points.reshape(-1, 3)
    all_points_flat_t = (all_points_flat + center).dot(np.linalg.inv(M.T))

    points_2d_proj_flat = cgroup.project(all_points_flat_t)
    points_2d_proj = points_2d_proj_flat.reshape(n_cams, n_joints, n_frames, 2)

    cam_names = cgroup.get_names()
    for cix, cname in enumerate(cam_names):
        offset = offsets_dict[cname]
        dx, dy = offset[0], offset[1]
        points_2d_proj[cix, :, :, 0] -= dx
        points_2d_proj[cix, :, :, 1] -= dy

    return scheme, bodyparts, points_2d_proj


def draw_projected_points(frames_2d, scheme, bodyparts, points):
    n_cams, n_joints, _ = points.shape
    out = []
    for cix, frame in enumerate(frames_2d):
        img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frame_out = label_frame(img, points[cix], scheme, bodyparts)
        img_out = cv2.cvtColor(frame_out, cv2.COLOR_RGB2BGR)
        out.append(img_out)
    return out

def visualize_combined(config, pose_fname, cgroup, offsets_dict,
                       fnames_2d, fname_3d, out_fname):

    should_load_3d = (cgroup is not None) and \
        (pose_fname is not None) and \
        (offsets_dict is not None)

    if should_load_3d:
        scheme, bodyparts, points_2d_proj = get_projected_points(config, pose_fname, cgroup, offsets_dict)

    # if angle_fname is not None:
    #     angles = pd.read_csv(angle_fname)
    #     bad_cols = ['fnum']
    #     ang_names = [col for col in angles.columns if col not in bad_cols]
    # else:
    ang_names = []
    angles = None

    ang_values = dict()

    for name in ang_names:
        vals = np.array(angles[name])
        angf = signal.medfilt(vals, kernel_size=5)
        err = np.abs(angf - vals)
        err[np.isnan(err)] = 10000

        vals[err > 10] = np.nan
        nans, ix = nan_helper(vals)
        # some data missing, but not too much
        if np.sum(nans) > 0 and np.sum(~nans) > 5:
            vals[nans] = np.interp(ix(nans), ix(~nans), vals[~nans])

        ang_values[name] = vals

    caps_2d = [cv2.VideoCapture(v) for v in fnames_2d]
    cap_3d = cv2.VideoCapture(fname_3d)

    pp = get_plotting_params(caps_2d, cap_3d, ang_names)
    nframes = pp['nframes']
    fps = pp['fps']
    start_img = get_start_image(pp, ang_names)

    ang_window_size = 100
    pad_size = ang_window_size

    ang_values_padded = dict()
    for name, angles in ang_values.items():
        ang_values_padded[name] = np.pad(angles, pad_size,
                                         mode='constant', constant_values=np.nan)

    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    writer = cv2.VideoWriter(out_fname, fourcc, round(fps, ndigits=2),
                             (pp['width_total'], pp['height_total']))

    q = queue.Queue(maxsize=50)

    thread = threading.Thread(target=write_frame_thread,
                              args=(writer, q))
    thread.start()

    for framenum in trange(nframes, ncols=70):
        ret, frames_2d, frame_3d = read_frames(caps_2d, cap_3d)
        if not ret:
            break

        if should_load_3d:
            frames_2d = draw_projected_points(
                frames_2d, scheme, bodyparts, points_2d_proj[:, :, framenum])

        all_angles = []
        for angnum, name in enumerate(ang_names):
            a = framenum + pad_size - ang_window_size//2
            b = a + ang_window_size
            angles = ang_values_padded[name][a:b]
            all_angles.append(angles)

        imout = draw_data(start_img, frames_2d, frame_3d, all_angles, pp)
        q.put(imout)

    for cap in caps_2d:
        cap.release()
    cap_3d.release()

    q.put(None)
    thread.join()
    writer.release()

def process_session(config, session_path):
    # filtered = config['filter']['enabled']
    # if filtered:
    #     pipeline_videos_labeled_2d = config['pipeline']['videos_labeled_2d_filter']
    # else:
    #     pipeline_videos_labeled_2d = config['pipeline']['videos_labeled_2d']
    pipeline_videos_labeled_3d = config['pipeline']['videos_labeled_3d']
    pipeline_videos_raw = config['pipeline']['videos_raw']
    # pipeline_angles = config['pipeline']['angles']

    if config['filter3d']['enabled']:
        pipeline_pose_3d = config['pipeline']['pose_3d_filter']
    else:
        pipeline_pose_3d = config['pipeline']['pose_3d']
    pipeline_videos_combined = config['pipeline']['videos_combined']

    video_ext = config['video_extension']

    vid_fnames_2d = glob(os.path.join(session_path,
                                      pipeline_videos_raw, "*."+video_ext))

    # vid_fnames_2d = glob(os.path.join(session_path,
    #                                   pipeline_videos_labeled_2d, "*.avi"))

    vid_fnames_3d = glob(os.path.join(session_path,
                                      pipeline_videos_labeled_3d, "*.mp4"))
    vid_fnames_3d = sorted(vid_fnames_3d, key=natural_keys)

    fnames_2d = defaultdict(list)
    for vid in vid_fnames_2d:
        vidname = get_video_name(config, vid)
        fnames_2d[vidname].append(vid)

    fnames_3d = defaultdict(list)
    for vid in vid_fnames_3d:
        vidname = true_basename(vid)
        fnames_3d[vidname].append(vid)

    cgroup = None
    calib_folder = find_calibration_folder(config, session_path)
    if calib_folder is not None:
        calib_fname = os.path.join(calib_folder,
                                   config['pipeline']['calibration_results'],
                                   'calibration.toml')
        if os.path.exists(calib_fname):
            cgroup = CameraGroup.load(calib_fname)

    # angle_fnames = glob(os.path.join(session_path,
    #                                  pipeline_angles, '*.csv'))
    # angle_fnames = sorted(angle_fnames, key=natural_keys)

    outdir = os.path.join(session_path, pipeline_videos_combined)

    if len(vid_fnames_3d) > 0:
        os.makedirs(outdir, exist_ok=True)

    for vid_fname in vid_fnames_3d:
        basename = true_basename(vid_fname)

        out_fname = os.path.join(outdir, basename+'.mp4')
        pose_fname = os.path.join(session_path, pipeline_pose_3d, basename+'.csv')

        if os.path.exists(out_fname) and \
           abs(get_nframes(out_fname) - get_nframes(vid_fname)) < 100:
            continue

        if not os.path.exists(pose_fname):
            print(out_fname, 'missing 3d data')
            continue

        if len(fnames_2d[basename]) == 0:
            print(out_fname, 'missing 2d videos')
            continue

        if len(fnames_3d[basename]) == 0:
            print(out_fname, 'missing 3d videos')
            continue

        fname_3d_current = fnames_3d[basename][0]
        fnames_2d_current = fnames_2d[basename]
        fnames_2d_current = sorted(fnames_2d_current, key=natural_keys)

        cam_names = [get_cam_name(config, fname) for fname in fnames_2d_current]

        print(out_fname)

        video_folder = os.path.join(session_path, pipeline_videos_raw)
        offsets_dict = load_offsets_dict(config, cam_names, video_folder)

        cgroup_subset = cgroup.subset_cameras_names(cam_names)

        visualize_combined(config, pose_fname, cgroup_subset, offsets_dict,
                           fnames_2d_current, fname_3d_current, out_fname)


label_combined_all = make_process_fun(process_session)



#!/usr/bin/env python3

import os.path

# Dependencies for video:
import os
from glob import glob
import io
from contextlib import redirect_stdout

from .common import natural_keys, make_process_fun

def rename_dlc_files(folder, base):
    files = glob(os.path.join(folder, base+'*'))
    for fname in files:
        basename = os.path.basename(fname)
        _, ext = os.path.splitext(basename)
        os.rename(os.path.join(folder, basename),
                  os.path.join(folder, base + ext))


def process_session(config, session_path):
    pipeline_videos_raw = config['pipeline']['videos_raw']
    pipeline_pose = config['pipeline']['pose_2d']
    video_ext = config['video_extension']

    config_name = os.path.join(config['model_folder'], 'config.yaml')

    source_folder = os.path.join(session_path, pipeline_videos_raw)
    outdir = os.path.join(session_path, pipeline_pose)

    videos = glob(os.path.join(source_folder, '*.'+video_ext))
    videos = sorted(videos, key=natural_keys)

    if len(videos) > 0:
        os.makedirs(outdir, exist_ok=True)

    videos_to_process = []
    for video in videos:
        basename = os.path.basename(video)
        basename, ext = os.path.splitext(basename)

        dataname = os.path.join(outdir, basename + '.h5')
        if os.path.exists(dataname):
            continue
        else:
            rename_dlc_files(outdir, basename) # try renaming in case it processed before
            if os.path.exists(dataname):
                print(video)
                continue
            else:
                videos_to_process.append(video)

    if len(videos_to_process) > 0:
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        import deeplabcut
        trap = io.StringIO()
        for i in range(0, len(videos_to_process), 5):
            batch = videos_to_process[i:i+5]
            for video in batch:
                print(video)
            with redirect_stdout(trap):
                deeplabcut.analyze_videos(config_name, batch,
                                          videotype=video_ext, save_as_csv=False,
                                          destfolder=outdir, TFGPUinference=False)
            for video in batch:
                basename = os.path.basename(video)
                basename, ext = os.path.splitext(basename)
                rename_dlc_files(outdir, basename)


pose_videos_all = make_process_fun(process_session)



#!/usr/bin/env python3

import os.path
import numpy as np
from glob import glob
import pandas as pd
import cv2
import skvideo.io
from tqdm import trange

from matplotlib.pyplot import get_cmap

from .common import make_process_fun, natural_keys, get_nframes

def connect(img, points, bps, bodyparts, col=(0,255,0,255)):
    try:
        ixs = [bodyparts.index(bp) for bp in bps]
    except ValueError:
        return

    for a, b in zip(ixs, ixs[1:]):
        if np.any(np.isnan(points[[a,b]])):
            continue
        pa = tuple(np.int32(points[a]))
        pb = tuple(np.int32(points[b]))
        cv2.line(img, tuple(pa), tuple(pb), col, 4)

def connect_all(img, points, scheme, bodyparts):
    cmap = get_cmap('tab10')
    for cnum, bps in enumerate(scheme):
        col = cmap(cnum % 10, bytes=True)
        col = [int(c) for c in col]
        connect(img, points, bps, bodyparts, col)


def label_frame(img, points, scheme, bodyparts, cmap='tab10'):
    n_joints, _ = points.shape

    cmap_c = get_cmap(cmap)
    connect_all(img, points, scheme, bodyparts)

    for lnum, (x, y) in enumerate(points):
        if np.isnan(x) or np.isnan(y):
            continue
        x = np.clip(x, 1, img.shape[1]-1)
        y = np.clip(y, 1, img.shape[0]-1)
        x = int(round(x))
        y = int(round(y))
        # col = cmap_c(lnum % 10, bytes=True)
        # col = [int(c) for c in col]
        col = (255, 255, 255)
        cv2.circle(img,(x,y), 7, col[:3], -1)

    return img

def visualize_labels(config, labels_fname, vid_fname, outname):

    try:
        scheme = config['labeling']['scheme']
    except KeyError:
        scheme = []

    if isinstance(labels_fname, str):
        dlabs = pd.read_hdf(labels_fname)
    elif isinstance(labels_fname, pd.DataFrame):
        dlabs = labels_fname
    else:
        raise TypeError('visualize_labels could not understand type for labels: {}', type(labels_fname))
        
    if len(dlabs.columns.levels) > 2:
        scorer = dlabs.columns.levels[0][0]
        dlabs = dlabs.loc[:, scorer]

    if len(scheme) == 0:
        bodyparts = list(dlabs.columns.levels[0])
    else:
        bodyparts = sorted(set([x for dx in scheme for x in dx]))

    cap = cv2.VideoCapture(vid_fname)
    # cap.set(1,0)

    fps = cap.get(cv2.CAP_PROP_FPS)

    writer = skvideo.io.FFmpegWriter(outname, inputdict={
        # '-hwaccel': 'auto',
        '-framerate': str(fps),
    }, outputdict={
        '-vcodec': 'h264', '-qp': '28',
        '-pix_fmt': 'yuv420p', # to support more players
        '-vf': 'pad=ceil(iw/2)*2:ceil(ih/2)*2' # to handle width/height not divisible by 2
    })

    last = len(dlabs)

    cmap = get_cmap('tab10')

    points = [(dlabs[bp]['x'], dlabs[bp]['y']) for bp in bodyparts]
    points = np.array(points)

    scores = [dlabs[bp]['likelihood'] for bp in bodyparts]
    scores = np.array(scores)
    scores[np.isnan(scores)] = 0
    scores[np.isnan(points[:, 0])] = 0

    good = np.array(scores) > 0.1
    points[:, 0, :][~good] = np.nan
    points[:, 1, :][~good] = np.nan

    all_points = points

    for ix in trange(last, ncols=70):
        ret, frame = cap.read()
        if not ret:
            break

        img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        points = all_points[:, :, ix]
        img = label_frame(img, points, scheme, bodyparts)

        writer.writeFrame(img)

    cap.release()
    writer.close()



def process_session(config, session_path, filtered=False):
    pipeline_videos_raw = config['pipeline']['videos_raw']
    if filtered:
        pipeline_videos_labeled = config['pipeline']['videos_labeled_2d_filter']
        pipeline_pose = config['pipeline']['pose_2d_filter']
    else:
        pipeline_videos_labeled = config['pipeline']['videos_labeled_2d']
        pipeline_pose = config['pipeline']['pose_2d']

    video_ext = config['video_extension']

    print(session_path)

    labels_fnames = glob(os.path.join(session_path, pipeline_pose, '*.h5'))
    labels_fnames = sorted(labels_fnames, key=natural_keys)

    outdir = os.path.join(session_path, pipeline_videos_labeled)

    if len(labels_fnames) > 0:
        os.makedirs(outdir, exist_ok=True)

    for fname in labels_fnames:
        basename = os.path.basename(fname)
        basename = os.path.splitext(basename)[0]

        out_fname = os.path.join(outdir, basename+'.mp4')
        vidname = os.path.join(session_path, pipeline_videos_raw, basename+'.'+video_ext)

        if os.path.exists(vidname):
            if os.path.exists(out_fname) and \
               abs(get_nframes(out_fname) - get_nframes(vidname)) < 100:
                continue
            print(out_fname)

            visualize_labels(config, fname, vidname, out_fname)


label_videos_all = make_process_fun(process_session, filtered=False)
label_videos_filtered_all = make_process_fun(process_session, filtered=True)



#!/usr/bin/env python3

import os
import os.path
import subprocess
from glob import glob
import sys
from collections import deque
import re
import cv2
from multiprocessing import Pool
from .common import process_all, get_video_params, natural_keys

if len(sys.argv) < 2:
    source_dir = os.getcwd()
else:
    source_dir = sys.argv[1]

source_dir = os.path.abspath(source_dir)


def same_length(vid1, vid2):
    params1 = get_video_params(vid1)
    params2 = get_video_params(vid2)
    return abs(params1['nframes'] - params2['nframes']) < 5


def process_video(fname, outname, encoding_params):
    # print(outname, 'started')
    if os.path.exists(outname) and same_length(fname, outname):
        return

    video_speed = encoding_params.get('converted_video_speed', 1)
    quality = encoding_params.get('video_quality', 28)
    gpu_enabled = encoding_params.get('gpu_enabled', False)

    params = get_video_params(fname)

    if video_speed != 1:
        vfilter = 'setpts={:.2f}*PTS, fps=fps={:.2f}, pad=ceil(iw/2)*2:ceil(ih/2)*2'.format(
            1.0/video_speed, params['fps']*video_speed
        )
    else:
        vfilter = 'pad=ceil(iw/2)*2:ceil(ih/2)*2'

    print(outname)
    if gpu_enabled:
        subprocess.run(['ffmpeg', '-y', '-i', fname,
                        '-hide_banner', '-loglevel', 'error', # '-stats',
                        '-vcodec', 'h264_nvenc',
                        '-rc', 'constqp', '-qp', str(quality),
                        '-pix_fmt', 'yuv420p',
                        '-filter:v', vfilter,
                        outname])
    else:
        subprocess.run(['ffmpeg', '-y', '-i', fname,
                        '-hide_banner', '-loglevel', 'error', # '-stats',
                        '-vcodec', 'h264', '-qp', str(quality), '-pix_fmt', 'yuv420p',
                        '-filter:v', vfilter,
                        outname])


    # print(outname, 'finished')

def process_folder(config, path):
    print(path)

    vidnames = glob(os.path.join(path, config['pipeline']['videos_raw'],
                                 '*.' + config['video_extension']))
    vidnames = sorted(vidnames, key=natural_keys)

    outpath = os.path.join(path, config['pipeline']['videos_raw_mp4'])

    os.makedirs(outpath, exist_ok=True)

    n_proc = config.get('encoding_nproc', 3)

    pool = Pool(n_proc)

    encoding_params = {
        'converted_video_speed': config['converted_video_speed'],
        'video_quality': config['video_quality'],
        'gpu_enabled': config['gpu_enabled']
    }

    for vidname in vidnames:
        basename = os.path.basename(vidname)
        base, ext = os.path.splitext(basename)
        outname = os.path.join(outpath, base+'.mp4')
        # if not (os.path.exists(outname) and same_length(vidname, outname)):
            # process_video(vidname, outname)
        pool.apply_async(process_video, (vidname, outname, encoding_params))

    pool.close()
    pool.join()

    return vidnames

def convert_all(config):

    process_all(config, process_folder)



#!/usr/bin/env python3
from flask import Flask
from flask import jsonify
from flask import request, send_from_directory
from flask_compress import Compress
from flask_ipban import IpBan
from werkzeug.utils import safe_join

from glob import glob
import os
from collections import deque, defaultdict
import re
import cv2
import string
import random

import pandas as pd
import numpy as np

from .anipose import load_config
from .common import find_calibration_folder, \
    get_video_name, get_cam_name, natural_keys, true_basename
from .project_2d import get_projected_points
from .triangulate import load_offsets_dict

from aniposelib.cameras import CameraGroup
import toml
import json

valid_tokens = set()

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    return [ atoi(c) for c in re.split('(\d+)', text) ]


SERVER_PASSWORD = os.environ.get('ANIPOSE_PASSWORD', 'password')

# creates a Flask application, named app
app = Flask(__name__)
# app.config['COMPRESS_LEVEL'] = 8
# app.config['COMPRESS_ALGORITHM'] = 'br'
Compress(app)
# squeeze = Squeeze()
# squeeze.init_app(app)

ip_ban = IpBan(ban_seconds=3600, ban_count=5, persist=True, ipc=True)
ip_ban.init_app(app)
ip_ban.load_nuisances()

def get_video_fnames(config, session_path):
    fnames = glob(safe_join(session_path, config['pipeline']['videos_raw_mp4'], '*.mp4'))
    return fnames

def get_folders(path):
    folders = next(os.walk(path))[1]
    return sorted(folders)

def generate_token(length): 
    letters = string.ascii_letters + '_'
    token = ''.join(random.choice(letters) for i in range(length))
    return token

def process_all(config, source_dir, process_session, **args):
    pipeline_prefix = source_dir

    output = dict()

    x = process_session(config, pipeline_prefix, **args)
    if x is not None:
        output[()] = x

    folders = get_folders(pipeline_prefix)
    level = 1

    q = deque()

    next_folders = [ (safe_join(pipeline_prefix, folder),
                      (folder,),
                      level)
                     for folder in folders ]
    q.extend(next_folders)

    while len(q) != 0:
        path, past_folders, level = q.popleft()

        x = process_session(config, path, **args)
        if x is not None:
            output[past_folders] = x

        folders = get_folders(path)
        next_folders = [ (safe_join(path, folder),
                          past_folders + (folder,),
                          level+1)
                         for folder in folders ]
        q.extend(next_folders)

    return output


def get_unique_behaviors(session_path):
    session = os.path.basename(session_path)
    path = safe_join(session_path, 'behaviors.json')
    if not os.path.exists(path):
        return [], {}

    with open(path) as json_file:
        behaviors = json.load(json_file)

    session_behaviors = set()
    trial_behaviors = {}
    folders = list(behaviors.keys())
    for folder in folders:
        filenames = list(behaviors[folder].keys())
        for file in filenames:
            unique_behaviors = {}
            rel_path = safe_join(session, folder, file)
            bouts = behaviors[folder][file]
            for key in list(bouts.keys()):
                behavior = bouts[key]['behavior']
                unique_behaviors[behavior] = True
                session_behaviors.add(behavior)
            trial_behaviors[rel_path] = unique_behaviors

    session_behaviors = list(session_behaviors)
    return session_behaviors, trial_behaviors


def get_config(session, append_prefix=True):
    if append_prefix:
        session_path = safe_join(prefix, session)
    else:
        session_path = session
    config_fname = safe_join(session_path, 'config.toml')
    config_fname = os.path.normpath(config_fname)
    config = load_config(config_fname)
    return config

def load_2d_projections(session, folders, fname):
    config = get_config(session)

    pipeline_calibration_videos = config.get('pipeline', {}).get('calibration_videos', 'calibration')
    search_path = os.path.normpath(safe_join(prefix, session, *folders))
    calib_folder = find_calibration_folder(config, search_path) 
    calib_fname = safe_join(calib_folder, pipeline_calibration_videos, "calibration.toml")
    cgroup = CameraGroup.load(os.path.normpath(calib_fname))

    offsets_dict = load_offsets_dict(config, cgroup.get_names())

    bodyparts, points_2d_proj, all_scores = get_projected_points(
        config, fname, cgroup, offsets_dict)

    cam_names = cgroup.get_names()

    points_2d_proj[~np.isfinite(points_2d_proj)] = 0
    points_2d_proj = np.int32(np.round(points_2d_proj))
    out = dict()
    for i, cname in enumerate(cam_names):
        out[cname] = points_2d_proj[i].swapaxes(0,1).tolist()

    return out

def get_structure(cdir):

    config_path = safe_join(cdir, 'config.toml')
    single_project = False

    if os.path.exists(config_path):
        prefix = os.path.dirname(cdir) 
        single_project = True
    else:
        (root, dirs, files) = next(os.walk(cdir))
        config_exists = False
        for d in dirs:
            if os.path.exists(safe_join(cdir, d, 'config.toml')):
                config_exists = True

        if config_exists: 
            prefix = cdir
        else:
            print('No Anipose project found. Exiting...') 
            exit()

    return prefix, single_project


# a route where we will display a welcome message via an HTML template
@app.route('/')
def root():
    return app.send_static_file('index.html')
    # return "hello"

@app.route('/get-sessions')
def get_sessions():
    sessions = []
    if single_project:
        sessions.append(os.path.basename(cdir))
    else: 
        (root, dirs, files) = next(os.walk(prefix))
        dirs = sorted(dirs, key=natural_keys)
        for folder in dirs:
            if os.path.exists(safe_join(prefix, folder, 'config.toml')):
                sessions.append(folder)

        sessions = sorted(sessions)

    return jsonify({
        'sessions': sessions
    })

@app.route('/pose3d/<session>/<folders>/<filename>')
def get_3d(session, folders, filename):
    config = get_config(session)
    folders = folders.split('|')
    path = safe_join(prefix, session, *folders)
    path = safe_join(path, 'pose-3d', filename + '.csv')
    path = os.path.normpath(path)
    data = pd.read_csv(path)

    cols = [x for x in data.columns if '_error' in x]
    # bodyparts = sorted([c.replace('_error', '') for c in cols])
    config = get_config(session)
    bodyparts = get_bodyparts_scheme(config['labeling']['scheme'])

    vecs = []
    for bp in bodyparts:
        vec = np.array(data[[bp+'_x', bp+'_y', bp+'_z']])
        error = np.array(data[bp+'_error'])
        if config['triangulation']['optim']:
            error[np.isnan(error)] = 0
        else:
            error[np.isnan(error)] = 1000
        vec[error > 50] = np.nan
        vecs.append(vec)

    vecs = np.array(vecs).swapaxes(0, 1)
    m = np.nanmean(vecs, axis = 0)
    # std = np.nanmean(np.nanstd(m, axis = 0))
    std = np.nanmedian(np.diff(np.nanpercentile(m, [25, 75], axis=0), axis=0))
    vecs = 0.3 * vecs / std
    
    cm = np.nanmean(np.nanmean(vecs, axis = 1), axis = 0)
    vecs = vecs - cm
    vecs[~np.isfinite(vecs)] = 0

    return jsonify(vecs.tolist())

@app.route('/pose2dproj/<session>/<folders>/<filename>')
def get_2d_proj(session, folders, filename):

    folders = folders.split('|')
    path = os.path.normpath(safe_join(prefix, session))
    fname = safe_join(path, *folders, 'pose-3d', filename + '.csv')

    projs = load_2d_projections(session, folders, fname)
    return jsonify(projs)

def get_bodyparts_scheme(scheme):
    bodyparts = []
    for bp_list in scheme:
        for bp in bp_list:
            if bp not in bodyparts:
                bodyparts.append(bp)
    return bodyparts

@app.route('/metadata/<session>')
def get_metadata(session):
    config = get_config(session)
    video_speed = config.get('converted_video_speed', 1)
    scheme = config['labeling']['scheme']

    bodyparts = get_bodyparts_scheme(scheme)
    kps = dict(zip(bodyparts, range(len(bodyparts))))

    ix = 0
    new_scheme = []
    for bp_list in scheme:
        kps_ix = np.zeros(len(bp_list), dtype = int)
        for j, bp in enumerate(bp_list):
            kps_ix[j] = kps[bp]
        new_scheme.append(kps_ix.tolist())

    metadata = {'video_speed': video_speed, 'scheme': new_scheme};

    return jsonify(metadata)

@app.route('/behavior/<session>/<folders>/<filename>')
def get_behaviors(session, folders, filename):

    session_path = safe_join(prefix, session)
    path = safe_join(session_path, 'behaviors.json')
    if not os.path.exists(path):
        return jsonify([])
    
    with open(path) as json_file:
        behavior_dict = json.load(json_file)

    behaviors = behavior_dict.get(folders, {}).get(filename, {})
    # behaviors = add_laser(behaviors, folders, filename)

    return jsonify(behaviors)

def merge_behavior_changes(behavior_changes):

    session_changes = defaultdict(list)
    for b in list(behavior_changes.keys()):
        changes = behavior_changes[b]
        session_changes[changes[0]['session']].extend(changes)

    for session in session_changes.keys():

        changes = session_changes[session]
        path = safe_join(prefix, session, 'behaviors.json')
        if os.path.exists(path):
            with open(path, 'r') as json_file:
                behavior_dict = json.load(json_file)
        else:
            behavior_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))

        for change in changes:
            if change['modification'] == 'added':
                bout = change['new']
                if bout['folders'] not in behavior_dict:
                    behavior_dict[bout['folders']] = dict()
                if bout['filename'] not in behavior_dict[bout['folders']]:
                    behavior_dict[bout['folders']][bout['filename']] = dict()
                behavior_dict[bout['folders']][bout['filename']][bout['bout_id']] = bout

            elif change['modification'] == 'removed':
                bout = change['old']
                try:
                    behavior_dict[bout['folders']][bout['filename']].pop(bout['bout_id'])
                except KeyError:
                    pass # was never there?

            else: # properties of an existing bout were edited (resized, translated, behavior name changed)
                bout = change['old']
                edits = change['new']
                for key in list(edits.keys()):
                    bout[key] = edits[key]
                if bout['folders'] not in behavior_dict:
                    behavior_dict[bout['folders']] = dict()
                if bout['filename'] not in behavior_dict[bout['folders']]:
                    behavior_dict[bout['folders']][bout['filename']] = dict()
                behavior_dict[bout['folders']][bout['filename']][bout['bout_id']] = bout

        with open(path, 'w') as json_file:
            json.dump(behavior_dict, json_file, indent = 4)

    message = 'behavior labels successfully updated' 
    return message

@app.route('/unlock-editing', methods=['POST'])
def authenticate():
    password_req = request.get_json()
    password = password_req['password']
    token = -1
    if password == SERVER_PASSWORD:
        token = generate_token(10)
        valid_tokens.add(token)
    valid = check_token(token)
    response = jsonify({'token': token, 'valid': valid})
    return response

@app.route('/get-token/<token>')
def check_existing_token(token):
    valid = token in valid_tokens
    return jsonify({'valid': valid})

def check_token(token):
    valid = token in valid_tokens
    return valid

@app.route('/update-behavior', methods=['POST'])
def update_behaviors():
    req_data = request.get_json()
    behavior_changes = req_data['allBehaviorChanges']
    token = req_data['token']
    valid = check_token(token)
    updated_behaviors = 'invalid token'
    if valid:   
        updated_behaviors = merge_behavior_changes(behavior_changes)
    return updated_behaviors

@app.route('/download-behavior/<session>')
def download_behaviors(session):
    session_path = safe_join(prefix, session)
    path = safe_join(session_path, 'behaviors.json')
    if not os.path.exists(path):
        return jsonify([])
    
    with open(path) as json_file:
        behaviors = json.load(json_file)

    return jsonify(behaviors)

@app.route('/video/<session>/<folders>/<filename>')
def get_video(session, folders, filename):
    print(session, folders, filename)
    config = get_config(session)
    folders = folders.split('|')
    path = safe_join(prefix, session, *folders)
    path = safe_join(path, config['pipeline']['videos_raw_mp4'])
    print(path, filename + '.mp4')
    return send_from_directory(path, filename + '.mp4')

@app.route('/video/<session>/<filename>')
def get_video_2(session, filename):
    return get_video(session, "", filename)
    # print(session, folders, filename)
    # config = get_config(session)
    # folders = folders.split('|')
    # path = safe_join(prefix, session, *folders)
    # path = safe_join(path, config['pipeline']['videos_raw_mp4'])
    # print(path, filename + '.mp4')
    # return send_from_directory(path, filename + '.mp4')

@app.route('/framerate/<session>/<folders>/<filename>')
def get_framerate(session, folders, filename):
    config = get_config(session)
    path = safe_join(prefix, session, folders.replace('|', '/'),
                     config['pipeline']['videos_raw_mp4'], filename + '.mp4')
    cap = cv2.VideoCapture(path)
    fps = cap.get(cv2.CAP_PROP_FPS)
    cap.release()
    print(path, fps)
    return jsonify(fps)

def group_by_trial(fnames, session):
    config = get_config(session)
    cam_videos = defaultdict(list)
    for fname in fnames:
        name = get_video_name(config, fname)
        cam_videos[name].append(fname)
    names = sorted(cam_videos.keys(), key=natural_keys)
    out = []
    for name in names:
        fnames = [true_basename(x) for x in cam_videos[name]]
        cnames = [get_cam_name(config, f) for f in cam_videos[name]]
        out.append({
            'vidname': name,
            'camnames': cnames,
            'files': fnames
        })
    return out

@app.route('/get-trials/<session>')
def get_trials(session):
    # session = request.args['session']
    path = safe_join(prefix, session)
    print(path)
    session_behaviors, trial_behaviors = get_unique_behaviors(path)
    config = get_config(session)
    fnames_dict = process_all(config, path, get_video_fnames)
    out = []
    for key, fnames in fnames_dict.items():
        if len(fnames) == 0:
            continue
        fnames = sorted(fnames, key=natural_keys)
        fnames = group_by_trial(fnames, session)
        d = {
            'folder': '|'.join(key),
            'files': fnames
        }
        out.append(d)

    out = sorted(out, key=lambda x: natural_keys(x['folder']))

    return jsonify({
        "session": session,
        "folders": out,
        "trialBehaviors": trial_behaviors,
        "sessionBehaviors": session_behaviors
    })


def run_server():
    global cdir, prefix, single_project
    cdir = os.getcwd()
    prefix, single_project = get_structure(cdir)
    app.run(debug=False, host="0.0.0.0", port=5000)
    # app.run(debug=False, threaded=False, processes=5, host="0.0.0.0", port=5000)

# run the application
if __name__ == "__main__":
    run_server()



#!/usr/bin/env python3

import cv2
# from cv2 import aruco
from tqdm import trange
import numpy as np
import os, os.path
from glob import glob
from collections import defaultdict
import pandas as pd

## TODO: rewrite this whole file with aniposelib

from .common import \
    get_calibration_board, get_board_type, \
    find_calibration_folder, make_process_fun, \
    get_cam_name, get_video_name, load_intrinsics, load_extrinsics
from .triangulate import triangulate_optim, triangulate_simple, \
    reprojection_error, reprojection_error_und
from .calibrate_extrinsics import detect_aruco, estimate_pose, fill_points

def expand_matrix(mtx):
    z = np.zeros((4,4))
    z[0:3,0:3] = mtx[0:3,0:3]
    z[3,3] = 1
    return z

def process_trig_errors(config, fname_dict, cam_intrinsics, extrinsics, skip=20):
    minlen = np.inf
    caps = dict()
    for cam_name, fname in fname_dict.items():
        cap = cv2.VideoCapture(fname)
        caps[cam_name] = cap
        length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        minlen = min(length, minlen)

    cam_names = sorted(fname_dict.keys())

    board = get_calibration_board(config)

    cam_mats = []
    cam_mats_dist = []
    for cname in cam_names:
        mat = np.array(extrinsics[cname])
        left = np.array(cam_intrinsics[cname]['camera_mat'])
        cam_mats.append(mat)
        cam_mats_dist.append(left)

    cam_mats = np.array(cam_mats)
    cam_mats_dist = np.array(cam_mats_dist)

    go = skip
    all_points = []
    framenums = []
    all_rvecs = []
    all_tvecs = []
    for framenum in trange(minlen, desc='detecting', ncols=70):
        row = []
        rvecs = []
        tvecs = []

        for cam_name in cam_names:
            intrinsics = cam_intrinsics[cam_name]
            cap = caps[cam_name]
            ret, frame = cap.read()

            if framenum % skip != 0 and go <= 0:
                continue

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            # corners, ids = detect_aruco(gray, intrinsics)
            detected, stuff = estimate_pose(gray, intrinsics, board)
            if detected:
                corners, ids, rvec, tvec = stuff
                rvec = rvec.flatten()
                tvec = tvec.flatten()
            else:
                corners = ids = None
                rvec = np.zeros(3)*np.nan
                tvec = np.zeros(3)*np.nan

            points = fill_points(corners, ids, board)
            points_flat = points.reshape(-1, 1, 2)
            points_new = cv2.undistortPoints(
                points_flat,
                np.array(intrinsics['camera_mat']),
                np.array(intrinsics['dist_coeff']))

            row.append(points_new.reshape(points.shape))
            rvecs.append(rvec)
            tvecs.append(tvec)

        if ~np.all(np.isnan(row)):
            all_points.append(row)
            all_tvecs.append(tvecs)
            all_rvecs.append(rvecs)
            framenums.append(framenum)
            go = skip

        go = max(0, go-1)

    all_points_raw = np.array(all_points)
    all_rvecs = np.array(all_rvecs)
    all_tvecs = np.array(all_tvecs)
    framenums = np.array(framenums)

    shape = all_points_raw.shape

    all_points_3d = np.zeros((shape[0], shape[2], 3))
    all_points_3d.fill(np.nan)

    num_cams = np.zeros((shape[0], shape[2]))
    num_cams.fill(np.nan)

    errors = np.zeros((shape[0], shape[2]))
    errors.fill(np.nan)

    for i in trange(all_points_raw.shape[0], desc='triangulating', ncols=70):
        for j in range(all_points_raw.shape[2]):
            pts = all_points_raw[i, :, j, :]
            good = ~np.isnan(pts[:, 0])
            if np.sum(good) >= 2:
                # p3d = triangulate_optim(pts, cam_mats)
                p3d = triangulate_simple(pts[good], cam_mats[good])
                all_points_3d[i, j] = p3d[:3]
                errors[i,j] = reprojection_error_und(p3d, pts[good], cam_mats[good], cam_mats_dist[good])
                num_cams[i,j] = np.sum(good)

    ## all_tvecs
    # framenum, camera num, axis

    dout = pd.DataFrame()
    for bp_num in range(shape[2]):
        bp = 'corner_{}'.format(bp_num)
        for ax_num, axis in enumerate(['x','y','z']):
            dout[bp + '_' + axis] = all_points_3d[:, bp_num, ax_num]
        dout[bp + '_error'] = errors[:, bp_num]
        dout[bp + '_ncams'] = num_cams[:, bp_num]

    for cam_num in range(shape[1]):
        cname = cam_names[cam_num]
        for ax_num, axis in enumerate(['x','y','z']):
            key = 'cam_{}_r{}'.format(cname, axis)
            dout[key] = all_rvecs[:, cam_num, ax_num]
            key = 'cam_{}_t{}'.format(cname, axis)
            dout[key] = all_tvecs[:, cam_num, ax_num]

    dout['fnum'] = framenums

    return dout



def process_session(config, session_path):
    # pipeline_videos_raw = config['pipeline']['videos_raw']
    pipeline_calibration_videos = config['pipeline']['calibration_videos']
    pipeline_calibration_results = config['pipeline']['calibration_results']

    calibration_path = find_calibration_folder(config, session_path)

    if calibration_path is None:
        return

    videos = glob(os.path.join(calibration_path,
                               pipeline_calibration_videos,
                               '*.avi'))
    videos = sorted(videos)

    cam_videos = defaultdict(list)

    cam_names = set()

    for vid in videos:
        name = get_video_name(config, vid)
        cam_videos[name].append(vid)
        cam_names.add(get_cam_name(config, vid))

    vid_names = cam_videos.keys()
    cam_names = sorted(cam_names)

    outdir = os.path.join(calibration_path, pipeline_calibration_results)
    os.makedirs(outdir, exist_ok=True)

    intrinsics = load_intrinsics(outdir, cam_names)
    extrinsics = load_extrinsics(outdir)

    fname_dicts = dict()
    for name in vid_names:
        fnames = cam_videos[name]
        cam_names = [get_cam_name(config, f) for f in fnames]
        fname_dict = dict(zip(cam_names, fnames))
        fname_dicts[name] = fname_dict

    for vidname, fd in fname_dicts.items():
        outname_base = vidname + '.csv'
        outname = os.path.join(outdir, outname_base)

        if os.path.exists(outname):
            continue

        print(outname)
        dout = process_trig_errors(config, fd, intrinsics, extrinsics)
        dout.to_csv(outname, index=False)


get_errors_all = make_process_fun(process_session)



#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
from tqdm import tqdm, trange
import sys
from collections import defaultdict
from pprint import pprint

from .common import process_all, true_basename, natural_keys, get_cam_name

def get_angle_fnames(config, session_path):
    fnames = glob(os.path.join(session_path,
                               config['pipeline']['angles'],
                               '*.csv'))
    return fnames

def get_pose3d_fnames(config, session_path):
    fnames = glob(os.path.join(session_path,
                               config['pipeline']['pose_3d'],
                               '*.csv'))
    return fnames

def get_pose3d_filtered_fnames(config, session_path):
    fnames = glob(os.path.join(session_path,
                               config['pipeline']['pose_3d_filter'],
                               '*.csv'))
    return fnames

def get_pose2d_fnames(config, session_path):
    fnames = glob(os.path.join(session_path,
                               config['pipeline']['pose_2d'],
                               '*.h5'))
    return fnames

def get_pose2d_filtered_fnames(config, session_path):
    fnames = glob(os.path.join(session_path,
                               config['pipeline']['pose_2d_filter'],
                               '*.h5'))
    return fnames

def make_summarize_fun(get_fnames_session, output_fname, h5=False):

    def summarize_fun(config):
        output = process_all(config, get_fnames_session)

        datas = []
        items = sorted(output.items())

        for key,fnames in tqdm(items, ncols=70, desc='sessions'):
            fnames = sorted(fnames, key=natural_keys)
            for fname in tqdm(fnames, ncols=70, desc='files'):
                if h5:
                    d = pd.read_hdf(fname)
                    scorer = d.columns.levels[0][0]
                    d = d[scorer]
                else:
                    d = pd.read_csv(fname)

                for num,foldername in enumerate(key, start=1):
                    k = 'folder_{}'.format(num)
                    d[k] = foldername

                d['filename'] = true_basename(fname)
                datas.append(d)

        dout = pd.concat(datas)
        dout['project'] = config['project']

        outdir = os.path.join(config['path'],
                              config['pipeline']['summaries'])

        os.makedirs(outdir, exist_ok=True)

        outname = os.path.join(outdir, output_fname)

        print('Saving output...')
        dout.to_csv(outname, index=False)

        if h5:
            basename = true_basename(outname)
            outfull_new = os.path.join(outdir, basename+'.h5')
            dout.to_hdf(outfull_new, 'df_with_missing', format='table', mode='w')

    return summarize_fun


summarize_angles = make_summarize_fun(get_angle_fnames, 'angles.csv', h5=False)
summarize_pose3d = make_summarize_fun(get_pose3d_fnames, 'pose_3d.csv', h5=False)
summarize_pose3d_filtered = make_summarize_fun(get_pose3d_fnames, 'pose_3d_filtered.csv', h5=False)

summarize_pose2d = make_summarize_fun(get_pose2d_fnames, 'pose_2d.csv', h5=True)
summarize_pose2d_filtered = make_summarize_fun(get_pose2d_filtered_fnames, 'pose_2d_filtered.csv', h5=True)

def summarize_errors(config):
    output = process_all(config, get_pose2d_filtered_fnames)

    rows = []
    items = sorted(output.items())

    for key,fnames in tqdm(items, ncols=70, desc='sessions'):
        fnames = sorted(fnames, key=natural_keys)
        for fname in tqdm(fnames, ncols=70, desc='files'):
            data = pd.read_hdf(fname)
            scorer = data.columns.levels[0][0]
            data = data[scorer]

            bp_index = data.columns.names.index('bodyparts')
            bodyparts = list(data.columns.levels[bp_index])

            rates_row = dict()
            for bp in bodyparts:
                rates_row[bp] = np.mean(data[bp]['interpolated'])

            for num,foldername in enumerate(key, start=1):
                k = 'folder_{}'.format(num)
                rates_row[k] = foldername

            rates_row['filename'] = true_basename(fname)
            rates_row['cam_name'] = get_cam_name(config, fname)
            rows.append(rates_row)

    dout = pd.DataFrame(rows)
    dout['project'] = config['project']

    outdir = os.path.join(config['path'],
                          config['pipeline']['summaries'])

    os.makedirs(outdir, exist_ok=True)

    outname = os.path.join(outdir, 'errors.csv')

    print('Saving output...')
    dout.to_csv(outname, index=False)



#!/usr/bin/env python3

import os
import os.path
import toml
import click

pass_config = click.make_pass_decorator(dict)

DEFAULT_CONFIG = {
    'video_extension': 'avi',
    'converted_video_speed': 1,
    'calibration': {
        'animal_calibration': False,
        'calibration_init': None,
        'fisheye': False
    },
    'manual_verification': {
        'manually_verify': False
    },
    'triangulation': {
        'ransac': False,
        'optim': False,
        'scale_smooth': 2,
        'scale_length': 2,
        'scale_length_weak': 1,
        'reproj_error_threshold': 5,
        'score_threshold': 0.8,
        'n_deriv_smooth': 3,
        'constraints': [],
        'constraints_weak': [],
        'optim_chunking': False,
        'optim_chunking_size': 1000
    },
    'pipeline': {
        'videos_raw': 'videos-raw',
        'videos_raw_mp4': 'videos-raw-mp4',
        'pose_2d': 'pose-2d',
        'pose_2d_filter': 'pose-2d-filtered',
        'pose_2d_projected': 'pose-2d-proj',
        'pose_3d': 'pose-3d',
        'pose_3d_filter': 'pose-3d-filtered',
        'videos_labeled_2d': 'videos-labeled',
        'videos_labeled_2d_filter': 'videos-labeled-filtered',
        'calibration_videos': 'calibration',
        'calibration_results': 'calibration',
        'videos_labeled_3d': 'videos-3d',
        'videos_labeled_3d_filter': 'videos-3d-filtered',
        'angles': 'angles',
        'summaries': 'summaries',
        'videos_combined': 'videos-combined',
        'videos_compare': 'videos-compare',
        'videos_2d_projected': 'videos-2d-proj',
    },
    'filter': {
        'enabled': False,
        'type': 'medfilt',
        'medfilt': 13,
        'offset_threshold': 25,
        'score_threshold': 0.05,
        'spline': True,
        'n_back': 5,
        'multiprocessing': False
    },
    'filter3d': {
        'enabled': False,
        'medfilt': 17,
        'offset_threshold': 15
    }
}

def full_path(path):
    path_user = os.path.expanduser(path)
    path_full = os.path.abspath(path_user)
    path_norm = os.path.normpath(path_full)
    return path_norm

def load_config(fname):
    if fname is None:
        fname = 'config.toml'

    if os.path.exists(fname):
        config = toml.load(fname)
    else:
        config = dict()

    # put in the defaults
    if 'path' not in config:
        if os.path.exists(fname) and os.path.dirname(fname) != '':
            config['path'] = os.path.dirname(fname)
        else:
            config['path'] = os.getcwd()

    config['path'] = full_path(config['path'])

    if 'project' not in config:
        config['project'] = os.path.basename(config['path'])

    for k, v in DEFAULT_CONFIG.items():
        if k not in config:
            config[k] = v
        elif isinstance(v, dict): # handle nested defaults
            for k2, v2 in v.items():
                if k2 not in config[k]:
                    config[k][k2] = v2

    return config

@click.group()
@click.version_option()
@click.option('--config', type=click.Path(exists=True, dir_okay=False),
              help='The config file to use instead of the default "config.toml" .')
@click.pass_context
def cli(ctx, config):
    ctx.obj = load_config(config)

@cli.command()
@pass_config
def calibrate(config):
    from .calibrate import calibrate_all
    click.echo('Calibrating...')
    calibrate_all(config)

@cli.command()
@pass_config
def calibration_errors(config):
    from .calibration_errors import get_errors_all
    click.echo('Getting all the calibration errors...')
    get_errors_all(config)

@cli.command()
@click.option('--scorer', type=str, help='scorer to use for labeled data')
@pass_config
def tracking_errors(config, scorer=None):
    from .tracking_errors import get_tracking_errors
    click.echo('Comparing tracking to labeled data...')
    get_tracking_errors(config, scorer)
    
@cli.command()
@pass_config
def analyze(config):
    from .pose_videos import pose_videos_all
    click.echo('Analyzing videos...')
    pose_videos_all(config)

@cli.command()
@pass_config
def filter(config):
    from .filter_pose import filter_pose_all
    click.echo('Filtering tracked points...')
    filter_pose_all(config)

@cli.command()
@pass_config
def filter_3d(config):
    from .filter_3d import filter_pose_3d_all
    click.echo('Filtering tracked points...')
    filter_pose_3d_all(config)

@cli.command()
@pass_config
def triangulate(config):
    from .triangulate import triangulate_all
    click.echo('Triangulating points...')
    triangulate_all(config)


@cli.command()
@pass_config
def angles(config):
    from .compute_angles import compute_angles_all
    click.echo('Computing angles...')
    compute_angles_all(config)

@cli.command()
@pass_config
def summarize_3d(config):
    from .summarize import summarize_angles, summarize_pose3d, summarize_pose3d_filtered
    click.echo('Summarizing angles...')
    summarize_angles(config)

    click.echo('Summarizing 3D pose...')
    summarize_pose3d(config)

    if config['filter3d']['enabled']:
        click.echo('Summarizing 3D pose filtered...')
        summarize_pose3d_filtered(config)


@cli.command()
@pass_config
def summarize_2d(config):
    from .summarize import summarize_pose2d
    click.echo('Summarizing pose 2d...')
    summarize_pose2d(config)


@cli.command()
@pass_config
def summarize_2d_filter(config):
    from .summarize import summarize_pose2d_filtered
    click.echo('Summarizing pose 2d filtered...')
    summarize_pose2d_filtered(config)


@cli.command()
@pass_config
def summarize_errors(config):
    from .summarize import summarize_errors
    click.echo('Summarizing errors...')
    summarize_errors(config)


@cli.command()
@click.option('--nframes', default=200, type=int, show_default=True)
@click.option('--mode', default='bad', type=str, show_default=True)
@click.option('--no-pred', is_flag=True)
@click.option('--scorer', default=None, type=str)
@click.option('--name', default=None, type=str)
@pass_config
def extract_frames(config, nframes=200, mode='bad', no_pred=False, scorer=None, name=None):
    from .extract_frames import extract_frames_picked, extract_frames_random
    click.echo('Extracting frames...')
    if no_pred:
        mode = 'random'
        extract_frames_random(config, nframes, name=name)
    else:
        extract_frames_picked(config, mode, nframes, scorer=scorer, name=name)


@cli.command()
@pass_config
def project_2d(config):
    from .project_2d import project_2d_all
    click.echo('Projecting 3D points back to 2D...')
    project_2d_all(config)


@cli.command()
@pass_config
def label_2d_proj(config):
    from .label_videos_proj import label_proj_all
    click.echo('Making 2D videos from 3D projections...')
    label_proj_all(config)
    
@cli.command()
@pass_config
def label_2d(config):
    from .label_videos import label_videos_all
    click.echo('Labeling videos in 2D...')
    label_videos_all(config)

@cli.command()
@pass_config
def label_2d_filter(config):
    from .label_videos import label_videos_filtered_all
    click.echo('Labeling videos in 2D...')
    label_videos_filtered_all(config)

@cli.command()
@pass_config
def label_3d(config):
    from .label_videos_3d import label_videos_3d_all
    click.echo('Labeling videos in 3D...')
    label_videos_3d_all(config)

@cli.command()
@pass_config
def label_3d_filter(config):
    from .label_videos_3d import label_videos_3d_filtered_all
    click.echo('Labeling videos in 3D...')
    label_videos_3d_filtered_all(config)

@cli.command()
@pass_config
def label_combined(config):
    from .label_combined import label_combined_all
    click.echo('Labeling combined videos...')
    label_combined_all(config)

@cli.command()
@pass_config
def label_filter_compare(config):
    from .label_filter_compare import label_filter_compare_all
    click.echo('Labeling videos to compare filtered vs raw tracking...')
    label_filter_compare_all(config)

@cli.command()
@pass_config
def draw_calibration(config):
    from .common import get_calibration_board_image
    import cv2
    click.echo('Drawing calibration board...')
    img = get_calibration_board_image(config)
    cv2.imwrite('calibration.png', img)

@cli.command()
@pass_config
def train_autoencoder(config):
    from .train_autoencoder import train_autoencoder
    click.echo('Training autoencoder...')
    train_autoencoder(config)


@cli.command()
@pass_config
def run_data(config):
    from .calibrate import calibrate_all
    from .pose_videos import pose_videos_all
    from .triangulate import triangulate_all
    from .compute_angles import compute_angles_all

    click.echo('Analyzing videos...')
    pose_videos_all(config)

    if config['filter']['enabled']:
        from .filter_pose import filter_pose_all
        click.echo('Filtering tracked points...')
        filter_pose_all(config)

    click.echo('Calibrating...')
    calibrate_all(config)

    click.echo('Triangulating points...')
    triangulate_all(config)

    if config['filter3d']['enabled']:
        from .filter_3d import filter_pose_3d_all
        click.echo('Filtering 3D points...')
        filter_pose_3d_all(config)
    
    click.echo('Computing angles...')
    compute_angles_all(config)


@cli.command()
@pass_config
def run_viz(config):
    from .label_videos import label_videos_filtered_all, label_videos_all
    from .label_videos_3d import label_videos_3d_all

    click.echo('Labeling videos in 2D...')
    if config['filter']['enabled']:
        label_videos_filtered_all(config)
    else:
        label_videos_all(config)
    click.echo('Labeling videos in 3D...')
    label_videos_3d_all(config)


@cli.command()
def visualizer():
    from .server import run_server
    run_server()

@cli.command()
@click.option('--nproc', default=3, type=int, show_default=True)
@click.option('--quality', default=28, type=int, show_default=True)
@click.option('--gpu', is_flag=True)
@pass_config
def convert_videos(config, nproc=3, gpu=False, quality=28):
    from .convert_videos import convert_all
    config['encoding_nproc'] = nproc
    config['gpu_enabled'] = gpu
    config['video_quality'] = quality
    convert_all(config)

@cli.command()
@pass_config
def run_all(config):
    from .calibrate import calibrate_all
    from .pose_videos import pose_videos_all
    from .triangulate import triangulate_all
    from .compute_angles import compute_angles_all

    click.echo('Analyzing videos...')
    pose_videos_all(config)

    if config['filter']['enabled']:
        from .filter_pose import filter_pose_all
        click.echo('Filtering tracked points...')
        filter_pose_all(config)

    click.echo('Calibrating...')
    calibrate_all(config)

    click.echo('Triangulating points...')
    triangulate_all(config)

    if config['filter3d']['enabled']:
        from .filter_3d import filter_pose_3d_all
        click.echo('Filtering 3D points...')
        filter_pose_3d_all(config)

    click.echo('Computing angles...')
    compute_angles_all(config)

    from .label_videos import label_videos_filtered_all, label_videos_all
    from .label_videos_3d import label_videos_3d_all
    from .label_combined import label_combined_all

    click.echo('Labeling videos in 2D...')
    if config['filter']['enabled']:
        label_videos_filtered_all(config)
    else:
        label_videos_all(config)

    click.echo('Labeling videos in 3D...')
    label_videos_3d_all(config)

    click.echo('Labeling combined videos...')
    label_combined_all(config)

if __name__ == '__main__':
    cli()



#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
import cv2
from tqdm import tqdm, trange
from collections import defaultdict
from scipy import signal
import queue
import threading

from aniposelib.cameras import CameraGroup

from .common import make_process_fun, get_nframes, \
    get_video_name, get_cam_name, \
    get_video_params, get_video_params_cap, \
    get_data_length, natural_keys, true_basename, find_calibration_folder, \
    nan_helper

from .triangulate import load_offsets_dict

from .label_videos import label_frame


def write_frame_thread(writer, q):
    while True:
        frame = q.get(block=True)
        if frame is None:
            return
        writer.write(frame)
        # writer.writeFrame(frame)

def turn_to_black(frame):
    frame = np.float32(frame)
    white = np.all(frame > 220, axis=2)
    frame[white] = [0,0,0]
    frame[~white] *= 1.5
    frame = np.clip(frame, 0, 255).astype('uint8')
    return frame


def read_frames(caps_2d):
    frames_2d = []
    for cap in caps_2d:
        ret, frame = cap.read()
        if not ret:
            return False, None
        frames_2d.append(frame)

    return ret, frames_2d


def get_plotting_params(caps_2d):
    params_2d = [get_video_params_cap(c) for c in caps_2d]
    height_2d = max([p['height'] for p in params_2d])
    widths_2d = [round(p['width'] * height_2d/p['height']) for p in params_2d]

    width_total = sum(widths_2d)
    height_total = height_2d * 3

    nframes = min([p['nframes'] for p in params_2d])

    fps = np.mean([p['fps'] for p in params_2d])

    d = {
        'height_2d': height_2d,
        'widths_2d': widths_2d,
        'nframes': nframes,
        'fps': fps,
        'width_total': width_total,
        'height_total': height_total,
    }

    return d

def get_start_image(pp):
    start_img = np.zeros((pp['height_total'], pp['width_total'], 3), dtype='uint8')
    start_img[:] = 0

    return start_img


def draw_data(start_img, frames_raw, frames_2d, frames_2d_filt, pp):
    height_2d = pp['height_2d']
    widths_2d = pp['widths_2d']

    width_total = pp['width_total']
    height_total = pp['height_total']

    frames_raw_resized = [cv2.resize(f, (w, height_2d))
                          for f, w in zip(frames_raw, widths_2d)]
    
    frames_2d_resized = [cv2.resize(f, (w, height_2d))
                         for f, w in zip(frames_2d, widths_2d)]

    frames_2d_filt_resized = [cv2.resize(f, (w, height_2d))
                              for f, w in zip(frames_2d_filt, widths_2d)]
    
    imout = np.copy(start_img)
    imout[0:height_2d] = np.hstack(frames_raw_resized)
    imout[height_2d:height_2d*2] = np.hstack(frames_2d_resized)
    imout[height_2d*2:height_2d*3] = np.hstack(frames_2d_filt_resized)

    return imout


def visualize_compare(config, fnames_raw, fnames_2d, fnames_2d_filt, out_fname):

    caps_raw = [cv2.VideoCapture(v) for v in fnames_raw]
    caps_2d = [cv2.VideoCapture(v) for v in fnames_2d]
    caps_2d_filt = [cv2.VideoCapture(v) for v in fnames_2d_filt]

    pp = get_plotting_params(caps_2d)
    nframes = pp['nframes']
    fps = pp['fps']
    start_img = get_start_image(pp)

    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    writer = cv2.VideoWriter(out_fname, fourcc, round(fps, ndigits=2),
                             (pp['width_total'], pp['height_total']))

    q = queue.Queue(maxsize=50)

    thread = threading.Thread(target=write_frame_thread,
                              args=(writer, q))
    thread.start()

    for framenum in trange(nframes, ncols=70):
        ret0, frames_raw = read_frames(caps_raw)
        ret1, frames_2d = read_frames(caps_2d)
        ret2, frames_2d_filt = read_frames(caps_2d_filt)
        if not all([ret0, ret1, ret2]):
            break

        imout = draw_data(start_img, frames_raw, frames_2d, frames_2d_filt, pp)
        q.put(imout)

    for cap in caps_raw:
        cap.release()
    for cap in caps_2d:
        cap.release()
    for cap in caps_2d_filt:
        cap.release()

    q.put(None)
    thread.join()
    writer.release()

def process_session(config, session_path):
    pipeline_videos_labeled_2d = config['pipeline']['videos_labeled_2d']
    pipeline_videos_labeled_2d_filter = config['pipeline']['videos_labeled_2d_filter']
    pipeline_videos_raw = config['pipeline']['videos_raw']
    pipeline_videos_compare = config['pipeline']['videos_compare']

    video_ext = config['video_extension']

    vid_fnames = glob(os.path.join(session_path, pipeline_videos_raw, "*." + video_ext))

    fnames_raw = defaultdict(list)
    for vid in vid_fnames:
        vidname = get_video_name(config, vid)
        fnames_raw[vidname].append(vid)

    outdir = os.path.join(session_path, pipeline_videos_compare)

    if len(vid_fnames) > 0:
        os.makedirs(outdir, exist_ok=True)

    for vidname in sorted(fnames_raw.keys(), key=natural_keys):
        out_fname = os.path.join(outdir, vidname+'.mp4')
        
        vids_raw = sorted(fnames_raw[vidname], key=natural_keys)
        vid_fname = vids_raw[0]
        
        if os.path.exists(out_fname) and \
           abs(get_nframes(out_fname) - get_nframes(vid_fname)) < 100:
            continue

        vids_2d = [os.path.join(session_path, pipeline_videos_labeled_2d,
                                true_basename(f) + '.mp4')
                   for f in vids_raw]

        vids_2d_filtered = [os.path.join(session_path,
                                         pipeline_videos_labeled_2d_filter,
                                         true_basename(f) + '.mp4')
                            for f in vids_raw]
        
        if not all([os.path.exists(f) for f in vids_2d]):
            print(out_fname, 'missing labeled 2d videos')
            continue

        if not all([os.path.exists(f) for f in vids_2d_filtered]):
            print(out_fname, 'missing labeled filtered 2d videos')
            continue

        print(out_fname)
        
        visualize_compare(config, vids_raw, vids_2d, vids_2d_filtered, out_fname)


label_filter_compare_all = make_process_fun(process_session)
